﻿/******************************************************************************\
  XNSNET 2005 - 2007 Автор: Иванов M.A. [Shum.er] 2005 - 2007 Версия кода: 069
\******************************************************************************/
namespace XNSNET.Collections {

	using System;

	/// <summary>
	/// Объектный класс связанного узла.
	/// </summary>
	public abstract class ObjectedNode {
	
    private int mCount;
    private ObjectedNode mFirst;
    private ObjectedNode mFinal;
		private ObjectedNode mList;
		private ObjectedNode mPrev;
		private ObjectedNode mNext;

		public delegate void Handler( ObjectedNode iList, ObjectedNode iItem );

		private event Handler mAdded;

		private event Handler mRemoved;

		public event Handler Added {
			add {
				mAdded += value;
			}
			remove {
				mAdded -= value;
			}
		}

		public event Handler Removed {
			add {
				mRemoved += value;
			}
			remove {
				mRemoved -= value;
			}
		}

		internal void RaiseAdded( ObjectedNode iItem ) {
			if ( mAdded != null )
				mAdded( this, iItem );
		}

		internal void RaiseRemoved( ObjectedNode iItem ) {
			if ( mRemoved != null )
				mRemoved( this, iItem );
		}

		protected ObjectedNode() {
      Clearing();
      Invalidate();
    }

    private void Clearing() {
      mCount = 0;
      mFirst = null;
      mFinal = null;
    }

		public bool IsCleared() {
			return ( mCount == 0 && mFirst == null && mFinal == null );
		}

    public int Count {
      get {
        return mCount;
      }
      private set {
        if ( value < 1 || value == mCount )
          throw new ArgumentOutOfRangeException();
        mCount = value;
      }
    }

		public ObjectedNode First {
      get {
        return mFirst;
      }
      private set {
        if ( value == null || value == mFirst )
          throw new ArgumentNullException();
				mFirst = value;
      }
    }

		public ObjectedNode Final {
      get {
        return mFinal;
      }
			private set {
        if ( value == null || value == mFinal )
          throw new ArgumentNullException();
				mFinal = value;
      }
    }

		public virtual void Clear() {
      if ( this.Count < 1 )
        return;
      lock ( this ) {
				ObjectedNode _Next = this.First;
        ObjectedNode _Item;
        while ( _Next != null ) {
          _Item = _Next;
          _Next = _Next.Next;
          _Item.Invalidate();
					_Item.OnRemoved( this );
        }
        this.Clearing();
      }
    }

		protected ObjectedNode AddFinal( ObjectedNode iItem ) {
			
      if ( iItem == null )
        throw new ArgumentNullException();
      if ( !iItem.IsInvalid() )
        throw new ArgumentException();
      lock ( this ) {
				iItem.List = this;
        if ( this.Count == 0 ) {
          this.First = iItem;
        } else {
					iItem.Prev = this.Final;
					this.Final.Next = iItem;
        }
        this.Final = iItem;
        this.Count++;
				iItem.OnAdded( this );
				return iItem;
      }
    }

		protected ObjectedNode AddFirst( ObjectedNode iItem ) {
			
      if ( iItem == null )
        throw new ArgumentNullException();
      if ( !iItem.IsInvalid() )
        throw new ArgumentException();
      lock ( this ) {
				iItem.List = this;
        if ( this.Count == 0 ) {
          this.Final = iItem;
        } else {
					iItem.Next = this.First;
					this.First.Prev = iItem;
        }
        this.First = iItem;
        this.Count++;
				iItem.OnAdded( this );
				return iItem;
      }
    }

		protected ObjectedNode AddAfter( ObjectedNode iTodo, ObjectedNode iItem ) {
			
      if ( iTodo == null || iItem == null )
        throw new ArgumentNullException();
      if ( iTodo.IsInvalid() || !iItem.IsInvalid() )
        throw new ArgumentException();
      lock ( this ) {
				iItem.List = this;
        iItem.Prev = iTodo;
        if ( iTodo.Next == null ) {
          this.Final = iItem;
        } else {
          iItem.Next = iTodo.Next;
					iTodo.Next.Prev = iItem;
        }
        iTodo.Next = iItem;
        this.Count++;
				iItem.OnAdded( this );
				return iItem;
      }
    }

		protected ObjectedNode AddBefore( ObjectedNode iTodo, ObjectedNode iItem ) {
			
      if ( iTodo == null || iItem == null )
        throw new ArgumentNullException();
      if ( iTodo.IsInvalid() || !iItem.IsInvalid() )
        throw new ArgumentException();
      lock ( this ) {
				iItem.List = this;
        iItem.Next = iTodo;
        if ( iTodo.Prev == null ) {
          this.First = iItem;
        } else {
          iItem.Next = iTodo;
          iTodo.Prev.Next = iItem;
        }
        iTodo.Prev = iItem;
        this.Count++;
				iItem.OnAdded( this );
				return iItem;
      }
    }

		public virtual ObjectedNode Remove( ObjectedNode iItem ) {

      if ( iItem == null )
        throw new ArgumentNullException();

			lock ( this ) {
				if ( this.Count == 1 ) {
					this.Clearing();
				} else {
					if ( iItem.Prev != null && iItem.Next != null ) {
						iItem.Prev.Next = iItem.Next;
						iItem.Next.Prev = iItem.Prev;
					} else if ( iItem.Prev == null ) {
						this.First = iItem.Next;
						iItem.Next.Prev = null;
					} else if ( iItem.Next == null ) {
						this.Final = iItem.Prev;
						iItem.Prev.Next = null;
					}
					Count--;
				}
				iItem.Invalidate();
				iItem.OnRemoved( this );
				return iItem;
			}

    }

		/// <summary>
		/// Возвращает экземпляр списка владеющего этим элементом.
		/// </summary>
		public ObjectedNode List {
			get {
				return mList;
			}
			private set {
				if ( mList != null || value == null )
					throw new ArgumentNullException();
				mList = value;
			}
		}

		/// <summary>
		/// Возвращает приведущий элемент в связанном списке.
		/// </summary>
		public ObjectedNode Prev {
			get {
				return mPrev;
			}
			private set {
				if ( mPrev == null && value == null )
					throw new ArgumentNullException();
				mPrev = value;
			}
		}

		/// <summary>
		/// Возвращает следующий элемент в связанном списке.
		/// </summary>
		public ObjectedNode Next {
			get {
				return mNext;
			}
			private set {
				if ( mNext == null && value == null )
					throw new ArgumentNullException();
				mNext = value;
			}
		}

		private void Invalidate() {
			mList = null;
			mPrev = null;
			mNext = null;
		}

		public bool IsInvalid() {
			return ( mList == null && mPrev == null && mNext == null );
		}

		/// <summary> Удаляет этот элемент из связанного узла. </summary>
		/// <returns> Возвращает собственный экземпляр в случае успеха. </returns>
		public virtual ObjectedNode Remove() {
			if ( mList != null ) {
				return mList.Remove( this );
			}
			return null;
		}

		/// <summary> Событие о добавлении этого элемента в связанный узел. </summary>
		/// <param name="iList"> Узел в который был добавлен этот элемент. </param>
		public virtual void OnAdded( ObjectedNode iList ) {
			iList.RaiseAdded( this );
		}

		/// <summary> Событие о удалении этого элемента из связанного узла. </summary>
		/// <param name="iList"> Узел в который был добавлен этот элемент. </param>
		public virtual void OnRemoved( ObjectedNode iList ) {
			iList.RaiseRemoved( this );
		}

		public ObjectedNode[] Array {
			get {
				lock ( this ) {
					ObjectedNode[] _Array = new ObjectedNode[ Count ];
					int _Index = 0;
					ObjectedNode _Item = this.First;
					while ( _Item != null ) {
						_Array[ _Index ] = _Item;
						_Item = _Item.Next;
						_Index++;
					}
					return _Array;
				}
			}
		}
    
	}

}