﻿/******************************************************************************\
  XNSNET 2005 - 2007 Автор: Иванов M.A. [Shum.er] 2005 - 2007 Версия кода: 069
\******************************************************************************/
namespace XNSNET.Collections {

	using System;

	/// <summary>
	/// Шаблонный класс списка связанных элементов.
	/// </summary>
	/// <typeparam name="LIST"> Шаблонный тип класса списка связанных элементов. </typeparam>
	/// <typeparam name="ITEM"> Шаблонный тип класса элемента связанного списка. </typeparam>
	public abstract class TemplatedList<LIST, ITEM>
		where LIST: TemplatedList<LIST, ITEM>
		where ITEM: TemplatedItem<LIST, ITEM> {

		public delegate void Handler( LIST iList, ITEM iItem );

		private event Handler mAdded;
		private event Handler mRemoved;

		public event Handler Added {
			add {
				mAdded += value;
			}
			remove {
				mAdded -= value;
			}
		}

		public event Handler Removed {
			add {
				mRemoved += value;
			}
			remove {
				mRemoved -= value;
			}
		}

		internal void RaiseAdded( ITEM iItem ) {
			if ( mAdded != null ) {
				mAdded( (LIST)this, iItem );
			}
		}
		
		internal void RaiseRemoved( ITEM iItem ) {
			if ( mRemoved != null ) {
				mRemoved( (LIST)this, iItem );
			}
		}
	
    private int mCount;
		private ITEM mFirst;
		private ITEM mFinal;

		protected TemplatedList() {
      Clearing();
    }

    private void Clearing() {
      mCount = 0;
      mFirst = null;
      mFinal = null;
    }

    public int Count {
      get {
        return mCount;
      }
      private set {
        if ( value < 1 || value == mCount )
          throw new ArgumentOutOfRangeException();
        mCount = value;
      }
    }

		public ITEM First {
      get {
        return mFirst;
      }
      private set {
        if ( value == null || value == mFirst )
          throw new ArgumentNullException();
        mFirst = value;
      }
    }

		public ITEM Final {
      get {
        return mFinal;
      }
      private set {
        if ( value == null || value == mFinal )
          throw new ArgumentNullException();
        mFinal = value;
      }
    }

		public virtual void Clear() {
			lock ( this ) {
				if ( this.Count < 1 )
					return;
				ITEM _Next = this.First;
				ITEM _Item;
        while ( _Next != null ) {
          _Item = _Next;
          _Next = _Next.Next;
          _Item.Invalidate();
          _Item.OnRemoved( (LIST)this );
        }
        this.Clearing();
      }
    }

		protected ITEM AddFinal( ITEM iItem ) {
		
      if ( iItem == null )
        throw new ArgumentNullException();
			if ( !iItem.IsInvalid() )
        throw new ArgumentException();

      lock ( this ) {
				iItem.List = (LIST)this;
        if ( this.Count == 0 ) {
          this.First = iItem;
        } else {
          iItem.Prev = this.Final;
          this.Final.Next = iItem;
        }
        this.Final = iItem;
				this.Count++;
				iItem.OnAdded( (LIST)this );
				return iItem;
      }
      
    }

		protected ITEM AddFirst( ITEM iItem ) {

      if ( iItem == null )
        throw new ArgumentNullException();
			if ( !iItem.IsInvalid() )
        throw new ArgumentException();

      lock ( this ) {
				iItem.List = (LIST)this;
        if ( this.Count == 0 ) {
          this.Final = iItem;
        } else {
          iItem.Next = this.First;
          this.First.Prev = iItem;
        }
        this.First = iItem;
				this.Count++;
				iItem.OnAdded( (LIST)this );
				return iItem;
      }
    }

		protected ITEM AddAfter( ITEM iTodo, ITEM iItem ) {

      if ( iTodo == null || iItem == null )
        throw new ArgumentNullException();
			if ( iTodo.IsInvalid() || !iItem.IsInvalid() )
        throw new ArgumentException();

      lock ( this ) {
				iItem.List = (LIST)this;
        iItem.Prev = iTodo;
        if ( iTodo.Next == null ) {
          this.Final = iItem;
        } else {
          iItem.Next = iTodo.Next;
          iTodo.Next.Prev = iItem;
        }
        iTodo.Next = iItem;
				this.Count++;
				iItem.OnAdded( (LIST)this );
				return iItem;
      }
    }

		protected ITEM AddBefore( ITEM iTodo, ITEM iItem ) {

      if ( iTodo == null || iItem == null )
        throw new ArgumentNullException();
			if ( iTodo.IsInvalid() || !iItem.IsInvalid() )
        throw new ArgumentException();

      lock ( this ) {
				iItem.List = (LIST)this;
        iItem.Next = iTodo;
        if ( iTodo.Prev == null ) {
          this.First = iItem;
        } else {
          iItem.Next = iTodo;
          iTodo.Prev.Next = iItem;
        }
        iTodo.Prev = iItem;
        this.Count++;
				iItem.OnAdded( (LIST)this );
				return iItem;
      }
    }

		public virtual ITEM Remove( ITEM iItem ) {

      if ( iItem == null )
        throw new ArgumentNullException();
      
			lock ( this ) {
				if ( this.Count == 1 ) {
					this.Clearing();
				} else {
					if ( iItem.Prev != null && iItem.Next != null ) {
						iItem.Prev.Next = iItem.Next;
						iItem.Next.Prev = iItem.Prev;
					} else if ( iItem.Prev == null ) {
						this.First = iItem.Next;
						iItem.Next.Prev = null;
					} else if ( iItem.Next == null ) {
						this.Final = iItem.Prev;
						iItem.Prev.Next = null;
					}
					Count--;
				}
				iItem.Invalidate();
				iItem.OnRemoved( (LIST)this );
				return iItem;
			}

    }

		public ITEM[] Array {
			get {
				lock ( this ) {
					ITEM[] _Array = new ITEM[ Count ];
					int _Index = 0;
					ITEM _Item = this.First;
					while ( _Item != null ) {
						_Array[ _Index ] = _Item;
						_Item = _Item.Next;
						_Index++;
					}
					return _Array;
				}
			}
		}

	}

}