﻿/******************************************************************************\
  XNSNET 2005 - 2007 Автор: Иванов M.A. [Shum.er] 2005 - 2007 Версия кода: 069
\******************************************************************************/
namespace XNSNET.Collections {

	using System;

	/// <summary>
	/// Шаблонный класс связанного узла.
	/// </summary>
	/// <typeparam name="NODE"> Шаблонный тип класса связанного узла. </typeparam>
	public abstract class TemplatedNode<NODE>
		where NODE: TemplatedNode<NODE> {
	
    private int mCount;
		private NODE mFirst;
		private NODE mFinal;

		private NODE mList;
		private NODE mPrev;
		private NODE mNext;

		public delegate void Handler( NODE iList, NODE iItem );

		private event Handler mAdded;

		private event Handler mRemoved;

		public event Handler Added {
			add {
				mAdded += value;
			}
			remove {
				mAdded -= value;
			}
		}

		public event Handler Removed {
			add {
				mRemoved += value;
			}
			remove {
				mRemoved -= value;
			}
		}

		internal void RaiseAdded( NODE iItem ) {
			if ( mAdded != null )
				mAdded( (NODE)this, iItem );
		}

		internal void RaiseRemoved( NODE iItem ) {
			if ( mRemoved != null )
				mRemoved( (NODE)this, iItem );
		}

		protected TemplatedNode() {
      Clearing();
      Invalidate();
    }

    private void Clearing() {
      mCount = 0;
      mFirst = null;
      mFinal = null;
    }

    public int Count {
      get {
        return mCount;
      }
      private set {
        if ( value < 1 || value == mCount )
          throw new ArgumentOutOfRangeException();
        mCount = value;
      }
    }

		public NODE First {
      get {
        return mFirst;
      }
      private set {
        if ( value == null || value == mFirst )
          throw new ArgumentNullException();
        mFirst = value;
      }
    }

		public NODE Final {
      get {
        return mFinal;
      }
      private set {
        if ( value == null || value == mFinal )
          throw new ArgumentNullException();
        mFinal = value;
      }
    }

		public virtual void Clear() {
      if ( this.Count < 1 )
        return;
      lock ( this ) {
				NODE _Next = this.First;
				NODE _Item;
        while ( _Next != null ) {
          _Item = _Next;
          _Next = _Next.Next;
          _Item.Invalidate();
          _Item.OnRemoved( (NODE)this );
        }
        this.Clearing();
      }
    }

		protected NODE AddFinal( NODE iItem ) {
      if ( iItem == null )
        throw new ArgumentNullException();
      if ( !iItem.IsValidate() )
        throw new ArgumentException();
      lock ( this ) {
				iItem.List = (NODE)this;
        if ( this.Count == 0 ) {
          this.First = iItem;
        } else {
          iItem.Prev = this.Final;
          this.Final.Next = iItem;
        }
        this.Final = iItem;
				this.Count++;
				iItem.OnAdded( (NODE)this );
				return iItem;
      }
    }

		protected NODE AddFirst( NODE iItem ) {

      if ( iItem == null )
        throw new ArgumentNullException();
      if ( !iItem.IsValidate() )
        throw new ArgumentException();

      lock ( this ) {
				iItem.List = (NODE)this;
        if ( this.Count == 0 ) {
          this.Final = iItem;
        } else {
          iItem.Next = this.First;
          this.First.Prev = iItem;
        }
        this.First = iItem;
				this.Count++;
				iItem.OnAdded( (NODE)this );
				return iItem;
      }
    }

		protected NODE AddAfter( NODE iTodo, NODE iItem ) {

      if ( iTodo == null || iItem == null )
        throw new ArgumentNullException();
      if ( iTodo.IsValidate() || !iItem.IsValidate() )
        throw new ArgumentException();
      
      lock ( this ) {
				iItem.List = (NODE)this;
        iItem.Prev = iTodo;
        if ( iTodo.Next == null ) {
          this.Final = iItem;
        } else {
          iItem.Next = iTodo.Next;
          iTodo.Next.Prev = iItem;
        }
        iTodo.Next = iItem;
				this.Count++;
				iItem.OnAdded( (NODE)this );
				return iItem;
      }
    }

		protected NODE AddBefore( NODE iTodo, NODE iItem ) {

      if ( iTodo == null || iItem == null )
        throw new ArgumentNullException();
      if ( iTodo.IsValidate() || !iItem.IsValidate() )
        throw new ArgumentException();

      lock ( this ) {
				iItem.List = (NODE)this;
        iItem.Next = iTodo;
        if ( iTodo.Prev == null ) {
          this.First = iItem;
        } else {
          iItem.Next = iTodo;
          iTodo.Prev.Next = iItem;
        }
        iTodo.Prev = iItem;
        this.Count++;
				iItem.OnAdded( (NODE)this );
				return iItem;
      }
    }

		public virtual NODE Remove( NODE iItem ) {

      if ( iItem == null )
        throw new ArgumentNullException();

			lock ( this ) {
				if ( this.Count == 1 ) {
					this.Clearing();
				} else {
					if ( iItem.Prev != null && iItem.Next != null ) {
						iItem.Prev.Next = iItem.Next;
						iItem.Next.Prev = iItem.Prev;
					} else if ( iItem.Prev == null ) {
						this.First = iItem.Next;
						iItem.Next.Prev = null;
					} else if ( iItem.Next == null ) {
						this.Final = iItem.Prev;
						iItem.Prev.Next = null;
					}
					Count--;
        }
				iItem.Invalidate();
				iItem.OnRemoved( (NODE)this );
				return iItem;
			}

    }

		/// <summary>
		/// Возвращает экземпляр списка владеющего этим элементом.
		/// </summary>
		public NODE List {
			get {
				return mList;
			}
			private set {
				if ( mList != null || value == null )
					throw new ArgumentNullException();
				mList = value;
			}
		}

		/// <summary>
		/// Возвращает приведущий элемент в связанном списке.
		/// </summary>
		public NODE Prev {
			get {
				return mPrev;
			}
			private set {
				if ( mPrev == null && value == null )
					throw new ArgumentNullException();
				mPrev = value;
			}
		}

		/// <summary>
		/// Возвращает следующий элемент в связанном списке.
		/// </summary>
		public NODE Next {
			get {
				return mNext;
			}
			private set {
				if ( mNext == null && value == null )
					throw new ArgumentNullException();
				mNext = value;
			}
		}

		private void Invalidate() {
			mList = null;
			mPrev = null;
			mNext = null;
		}

		private bool IsValidate() {
			return ( mList == null && mPrev == null && mNext == null );
		}

		/// <summary> Удаляет этот элемент из связанного узла. </summary>
		/// <returns> Возвращает собственный экземпляр в случае успеха. </returns>
		public virtual NODE Remove() {
			if ( mList != null ) {
				return mList.Remove( (NODE)this );
			}
			return null;
		}

		/// <summary> Событие о добавлении этого элемента в связанный узел. </summary>
		/// <param name="iList"> Узел в который был добавлен этот элемент. </param>
		public virtual void OnAdded( NODE iList ) {
			iList.RaiseAdded( (NODE)this );
		}

		/// <summary> Событие о удалении этого элемента из связанного узла. </summary>
		/// <param name="iList"> Узел в который был добавлен этот элемент. </param>
		public virtual void OnRemoved( NODE iList ) {
			iList.RaiseRemoved( (NODE)this );
		}

		public NODE[] Array {
			get {
				lock ( this ) {
					NODE[] _Array = new NODE[ Count ];
					int _Index = 0;
					NODE _Item = this.First;
					while ( _Item != null ) {
						_Array[ _Index ] = _Item;
						_Item = _Item.Next;
						_Index++;
					}
					return _Array;
				}
			}
		}

	}
	
}